unit Reports;

interface

uses
  SysUtils, Classes, Forms, DB, DBTables, Dialogs, EXLReport, EXLReportExcelTLB, EXLReportBand,
  ActiveX, Graphics;

type
  TReportContainer = class(TDataModule)
    QuickStart: TEXLReport;
    MasterTable: TTable;
    DetailTable: TTable;
    DetailSource: TDataSource;
    MasterSource: TDataSource;
    UserFields: TEXLReport;
    FormatPro: TEXLReport;
    FormatCond: TEXLReport;
    RowHeight: TEXLReport;
    LabelReport: TEXLReport;
    CellData: TEXLReport;
    SumReport: TEXLReport;
    Totals: TEXLReport;
    DataCalc: TEXLReport;
    MasterDetail: TEXLReport;
    Groupping: TEXLReport;
    UserData: TEXLReport;
    SheetProp: TEXLReport;
    MultiSheet2: TEXLReport;
    MultiSheet: TEXLReport;
    procedure DataModuleCreate(Sender: TObject);
    procedure DataModuleDestroy(Sender: TObject);
    procedure UserFieldsBeforeBuild(Sender: TObject);
    procedure UserFieldsGetFieldValue(Sender: TObject; const Field: String;
      var Value: OleVariant);
    procedure FormatProFormatCell(Sender: TObject; Band: TEXLReportBand;
      RowInTemplate, RowInReport, ColumnInReport: Integer;
      Report: _Worksheet; const CellValue: OleVariant);
    procedure RowHeightFormatCell(Sender: TObject; Band: TEXLReportBand;
      RowInTemplate, RowInReport, ColumnInReport: Integer;
      Report: _Worksheet; const CellValue: OleVariant);
    procedure CellDataSetCellValue(Sender: TObject; Band: TEXLReportBand;
      RowInTemplate, RowInReport, ColumnInReport: Integer;
      var CellValue: OleVariant);
    procedure TotalsFormatCell(Sender: TObject; Band: TEXLReportBand;
      RowInTemplate, RowInReport, ColumnInReport: Integer;
      Report: _Worksheet; const CellValue: OleVariant);
    procedure TotalsBeforeBuild(Sender: TObject);
    procedure DataCalcFormatCell(Sender: TObject; Band: TEXLReportBand;
      RowInTemplate, RowInReport, ColumnInReport: Integer;
      Report: _Worksheet; const CellValue: OleVariant);
    procedure GrouppingBandsGroupHeader1CalcExpression(Sender: TEXLReportBand;
      var GroupValue: String);
    procedure GrouppingBeforeBuild(Sender: TObject);
    procedure UserDataDataEof(Sender: TObject; var Eof: Boolean);
    procedure UserDataDataFirst(Sender: TObject);
    procedure UserDataDataNext(Sender: TObject);
    procedure UserDataDataPrior(Sender: TObject);
    procedure UserDataGetFieldValue(Sender: TObject; const Field: String;
      var Value: OleVariant);
    procedure SheetPropAfterBuild(Sender: TObject; Report: _Worksheet;
      RowCount, ColCount: Integer);
  private
    { Private declarations }
    FDataFile: TStringList;
    FDFMFile: TStringList;
    FPASFile: TStringList;
    FCounter: Integer;
    FStartRow: Integer;
    FEndRow: Integer;
    function GetRecVal(const Str: String; var MT, DT: String): Integer;
    procedure GetTableNames(Num: Integer; var MT, DT: String; Comment: TStrings);
    procedure GetDFMText(const CompName: String; List: TStrings);
    procedure GetEventCode(const CompName: String; List: TStrings);
    procedure LoadMemos;
  public
    { Public declarations }
    function LoadReportData(const ReportName: String; Comment, DFMText,
      EventCode: TStrings): TEXLReport;
  end;

var
  ReportContainer: TReportContainer;

implementation

{$R *.dfm}

const
  _DataFileName = '\DemoData.txt';
  _DFMFileName = '\reports.dfm';
  _PASFileName = '\reports.pas';
  _Sign = '<<';
  _Comment = '//';
  _Msg = 'Cannot load a file "%s". %s error: %s';

function GetFirstPart(const Str: String): String;
var
  I: Integer;
begin
  I := Pos(':', Str);
  if I > 0 then Result := Copy(Str, 1, I - 1)
  else Result := Str;
end;

function GetLastPart(const Str: String): String;
var
  I: Integer;
begin
  I := Pos(':', Str);
  if I > 0 then Result := Copy(Str, I + 1, Length(Str))
  else Result := '';
end;

{ TReportContainer }

function TReportContainer.LoadReportData(const ReportName: String; Comment, DFMText,
  EventCode: TStrings): TEXLReport;

  procedure LoadCodes(Cmp: TEXLReport);
  begin
    while Assigned(Cmp) do
    begin
      GetDFMText(Cmp.Name, DFMText);
      GetEventCode(Cmp.Name, EventCode);
      Cmp := Cmp.SubReport;
    end;
  end;

var
  Cmp: TEXLReport;
  I: Integer;
  M, D: String;
begin
  MasterTable.Close;
  DetailTable.Close;
  Result := nil;
  EventCode.Clear;
  DFMText.Clear;
  Comment.Clear;
  for I := 0 to ComponentCount - 1 do
    if Components[I] is TEXLReport then
    begin
      Cmp := TEXLReport(Components[I]);
      if (Cmp.Title = ReportName) and (Cmp.Tag > 0) then
      begin
        Result := Cmp;
        Break;
      end;
    end;
  if Assigned(Result) then
  begin
    GetTableNames(Result.Tag, M, D, Comment);
    MasterTable.TableName := GetFirstPart(M);
    MasterTable.IndexName := GetLastPart(M);
    DetailTable.TableName := GetFirstPart(D);
    DetailTable.MasterFields := GetLastPart(D);
    DetailTable.IndexName := DetailTable.MasterFields;
    LoadCodes(Result);
    if M <> '' then MasterTable.Open;
    if D <> '' then DetailTable.Open;
  end
  else raise Exception.CreateFmt('"%s" report is not found.', [ReportName]);
end;

procedure TReportContainer.DataModuleCreate(Sender: TObject);
begin
  FDataFile := TStringList.Create;
  FDFMFile := TStringList.Create;
  FPASFile := TStringList.Create;
  LoadMemos;
end;

procedure TReportContainer.DataModuleDestroy(Sender: TObject);
begin
  FDataFile.Free;
  FDFMFile.Free;
  FPASFile.Free;
  ReportContainer := nil;
end;

function TReportContainer.GetRecVal(const Str: String; var MT, DT: String): Integer;
var
  I, N: Integer;
  S: String;
begin
  Result := - 1;
  N := 0;
  MT := '';
  DT := '';
  for I := Length(_Sign) + 1 to Length(Str) do
    if not (Str[I] in ['0'..'9']) then
    begin
      N := I - Length(_Sign) - 1;
      Break;
    end;
  if N <> 0 then
  begin
    Result := StrToInt(Copy(Str, Length(_Sign) + 1, N));
    N := Pos(' ', Str);
    S := Trim(Copy(Str, N + 1, Length(Str)));
    if Copy(S, 1, Length(_Comment)) <> _Comment then
    begin
      N := Pos(' ', S);
      if N > 0 then
      begin
        MT := Copy(S, 1, N - 1);
        DT := Trim(Copy(S, N + 1, Length(S)));
        if Copy(DT, 1, Length(_Comment)) = _Comment then DT := ''
        else begin
          N := Pos(' ', DT);
          if N > 0 then DT := Copy(DT, 1, N - 1);
        end;
      end
      else MT := S;
    end;
  end;
end;

procedure TReportContainer.GetTableNames(Num: Integer; var MT, DT: String; Comment: TStrings);
var
  I: Integer;
  S, M, D: String;
  Found: Boolean;
begin
  Found := False;
  for I := 0 to FDataFile.Count - 1 do
  begin
    S := FDataFile.Strings[I];
    if Found then
      if Copy(S, 1, Length(_Sign)) = _Sign then Break
      else Comment.Add(S)
    else
      if Copy(S, 1, Length(_Sign)) = _Sign then
        if GetRecVal(S, M, D) = Num then
        begin
          MT := M;
          DT := D;
          Found := True;
        end;
  end;
  if not Found then
    raise Exception.CreateFmt('Record of number %d does not exist in the data file.', [Num]);
end;

procedure TReportContainer.GetDFMText(const CompName: String;
  List: TStrings);
const
  _End = '  end';
  _Object = 'object';
  _EXLReport = 'TEXLReport';
var
  I: Integer;
  Found: Boolean;
  S: String;
begin
  Found := False;
  for I := 0 to FDFMFile.Count - 1 do
  begin
    S := FDFMFile.Strings[I];
    if Found then
    begin
      List.Add(S);
      if S = _End then Break;
    end
    else begin
//      Found := (Pos(_Object, S) > 0) and (Pos(CompName, S) > 0) and (Pos(_EXLReport, S) > 0);
      Found := (Pos(_Object, S) > 0) and (Pos(' ' + CompName + ':', S) > 0) and
        (Pos(_EXLReport, S) > 0);
      if Found then List.Add(S);
    end;
  end;
end;

procedure TReportContainer.GetEventCode(const CompName: String;
  List: TStrings);
const
  _End = 'end;';
  _Proc = 'procedure TReportContainer.';
var
  I: Integer;
  Found: Boolean;
  S, T: String;
begin
  Found := False;
  for I := 0 to FPASFile.Count - 1 do
  begin
    S := FPASFile.Strings[I];
    if Found then
    begin
      List.Add(S);
      if S = _End then
      begin
        Found := False;
        List.Add('');
      end;
    end
    else begin
      T := _Proc + CompName;
      Found := Copy(S, 1, Length(T)) = T;
      if Found then List.Add(S);
    end;
  end;
end;

procedure TReportContainer.LoadMemos;
var
  S: String;
begin
  S := ExtractFileDir(Application.ExeName);

  try
    FDataFile.LoadFromFile(S + _DataFileName);
  except
    on E: Exception do
      MessageDlg(Format(_Msg, [S + _DataFileName, E.ClassName, E.Message]), mtError, [mbOK], 0);
  end;

  try
    FDFMFile.LoadFromFile(S + _DFMFileName);
  except
    on E: Exception do
      MessageDlg(Format(_Msg, [S + _DFMFileName, E.ClassName, E.Message]), mtError, [mbOK], 0);
  end;

  try
    FPASFile.LoadFromFile(S + _PASFileName);
  except
    on E: Exception do
      MessageDlg(Format(_Msg, [S + _PASFileName, E.ClassName, E.Message]), mtError, [mbOK], 0);
  end;
end;

procedure TReportContainer.UserFieldsBeforeBuild(Sender: TObject);
begin
  FCounter := 1;
end;

procedure TReportContainer.UserFieldsGetFieldValue(Sender: TObject;
  const Field: String; var Value: OleVariant);
begin
  if AnsiCompareText(Field, 'No') = 0 then
  begin
    Value := FCounter;
    Inc(FCounter);
  end
  else if AnsiCompareText(Field, 'Name') = 0 then
    Value := MasterTable.FieldByName('LastName').AsString + ', ' +
             MasterTable.FieldByName('FirstName').AsString;
end;

procedure TReportContainer.FormatProFormatCell(Sender: TObject;
  Band: TEXLReportBand; RowInTemplate, RowInReport, ColumnInReport: Integer;
  Report: _Worksheet; const CellValue: OleVariant);
var
  R1, R2: String;
begin
  if (Band.BandType = xlrbtMasterData) and (ColumnInReport = 4) then
  begin
    R1 := 'A' + IntToStr(RowInReport);
    R2 := 'D' + IntToStr(RowInReport);
    if CellValue >= 45000 then
    begin
      Report.Range[R2, R2].Font.Color := clRed;
      Report.Range[R2, R2].Font.Bold := True;
    end
    else if CellValue <= 20000 then
    begin
      Report.Range[R1, R2].Font.Color := clWhite;
      Report.Range[R1, R2].Interior.Color := clRed;
    end;
  end;
end;

procedure TReportContainer.RowHeightFormatCell(Sender: TObject;
  Band: TEXLReportBand; RowInTemplate, RowInReport, ColumnInReport: Integer;
  Report: _Worksheet; const CellValue: OleVariant);
var
  S: String;
begin
  if Band.BandType = xlrbtMasterData then
  begin
    S := 'A' + IntToStr(RowInReport);
    if Report.Range[S, S].RowHeight < 21 then
      Report.Range[S, S].RowHeight := 21;
  end;
end;

procedure TReportContainer.CellDataSetCellValue(Sender: TObject;
  Band: TEXLReportBand; RowInTemplate, RowInReport, ColumnInReport: Integer;
  var CellValue: OleVariant);
begin
  if Band.BandType = xlrbtMasterData then
    CellValue := AnsiUpperCase(CellValue);
end;

procedure TReportContainer.TotalsFormatCell(Sender: TObject;
  Band: TEXLReportBand; RowInTemplate, RowInReport, ColumnInReport: Integer;
  Report: _Worksheet; const CellValue: OleVariant);
var
  S: String;
begin
  if Band.BandType = xlrbtMasterData then
  begin
    if FStartRow = 0 then FStartRow := RowInReport;
    FEndRow := RowInReport;
  end
  else if (Band.BandType = xlrbtSummary) and (ColumnInReport = 3)
    and (RowInTemplate >= 8) then
  begin
    S := 'C' + IntToStr(RowInReport);
    if RowInTemplate = 8 then
      Report.Range[S, S].Formula := Format('=Mode(C%d:C%d)', [FStartRow, FEndRow])
    else
      Report.Range[S, S].Formula := Format('=Median(C%d:C%d)', [FStartRow, FEndRow]);
  end;
end;

procedure TReportContainer.TotalsBeforeBuild(Sender: TObject);
begin
  FStartRow := 0;
  FEndRow := 0;
end;

procedure TReportContainer.DataCalcFormatCell(Sender: TObject;
  Band: TEXLReportBand; RowInTemplate, RowInReport, ColumnInReport: Integer;
  Report: _Worksheet; const CellValue: OleVariant);
begin
  if (Band.BandType = xlrbtSummary) and (ColumnInReport = 3)
     and (RowInTemplate = 5) then
    Report.Range['C2', 'C2'].Formula := '= C' + IntToStr(RowInReport); 
end;

procedure TReportContainer.GrouppingBandsGroupHeader1CalcExpression(
  Sender: TEXLReportBand; var GroupValue: String);
begin
  GroupValue := FormatDateTime('yyyymm', MasterTable.FieldByName('Event_date').AsDateTime);
end;

procedure TReportContainer.GrouppingBeforeBuild(Sender: TObject);
begin
  // Add an index for the groupping
  try
    MasterTable.Close;
    MasterTable.Exclusive := True;
    MasterTable.Open;
    MasterTable.AddIndex('ByDate', 'Event_date;Event_name', [ixCaseInsensitive]);
  except
  end;
  MasterTable.Close;
  MasterTable.Exclusive := False;
  MasterTable.Open;
  try
    MasterTable.IndexName := 'ByDate';
  except
  end;
end;

procedure TReportContainer.UserDataDataEof(Sender: TObject;
  var Eof: Boolean);
begin
  Eof := FCounter > 20;
end;

procedure TReportContainer.UserDataDataFirst(Sender: TObject);
begin
  Randomize;
  FCounter := 1;
end;

procedure TReportContainer.UserDataDataNext(Sender: TObject);
begin
  Inc(FCounter);
end;

procedure TReportContainer.UserDataDataPrior(Sender: TObject);
begin
  Dec(FCounter);
end;

procedure TReportContainer.UserDataGetFieldValue(Sender: TObject;
  const Field: String; var Value: OleVariant);
begin
  if AnsiCompareText(Field, 'No') = 0 then
    Value := FCounter
  else if AnsiCompareText(Field, 'String') = 0 then
    Value := Format('This is the string %d.', [FCounter])
  else if AnsiCompareText(Field, 'Number') = 0 then
    Value := Random(100);
end;

procedure TReportContainer.SheetPropAfterBuild(Sender: TObject;
  Report: _Worksheet; RowCount, ColCount: Integer);
var
  XLApp: Variant;
begin
  Report.Name := 'Sheet name';
  XLApp := Report.Application;
  XLApp.Run('UserMacro');  // Not to set all the procedure parameters.
end;

end.
